""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
from nloed import Model
from nloed import Design

import re as re
import copy as cp
import matplotlib.pyplot as plt

####################################################################################################
# SET UP RHS OF ODE
####################################################################################################

# state variales
y = cs.SX.sym('y',4)
# parameters
p = cs.SX.sym('p',8)

#symbolic RHS
rhs = cs.vertcat(y[0]*( cs.exp(p[0])*(y[2]/(cs.exp(p[1])+y[2])) - cs.exp(p[2])*y[1] ),
                 cs.exp(p[3])*(y[3]/(cs.exp(p[4])+y[3]))*y[1],
                 -cs.exp(p[5])*(y[2]/(cs.exp(p[1])+y[2]))*y[0],
                 y[1]*( cs.exp(p[6])*y[0] - cs.exp(p[7])*(y[3]/(cs.exp(p[4])+y[3])) ) )
#casadi RHS function
ode = cs.Function('ode',[y,p],[rhs])

####################################################################################################
# DEFINE RK-4 STEP ALGEBRA
####################################################################################################

#time step size
dt = 0.01
# Create symbolics for RK4 integration, as shown in Casadi examples
k1 = ode(y, p)
k2 = ode(y + dt/2.0*k1, p)
k3 = ode(y + dt/2.0*k2, p)
k4 = ode(y + dt*k3, p)
y_step = y + dt/6.0 * (k1 + 2*k2 + 2*k3 + k4)
# Create a function to perform one step of the RK integration
step = cs.Function('step',[y, p],[y_step])

####################################################################################################
# DEFINE TIME DISCRETIZATION, AND CONSTRUCT DETERMINISTIC MODEL VIA INTEGRATION
####################################################################################################

#define a sample pattern to apply in each control interval
sample_pattern= [500,500,500,500,500] 

# IC symbols (considered as controls here)
u = cs.SX.sym('u',2)
y0 = cs.vertcat(u,cs.SX([5,0]))

# create list to store symbols for each sample point
sample_list=[]
# create list to store times of each sample (for naming observ. vars)
times=[]

#Loop over the ste function to create symbolics for integrating across a sample interval
y_sample = y0
cntr=0
# loop over sample pattern
for num_stps in sample_pattern:
    #iterate steps indicated by pattern
    for k in range(num_stps):
        #propogate the state variables
        y_sample = step(y_sample, p)
        cntr+=1
    #save the state symbols and times of each sample
    sample_list.append(y_sample)
    times.append(cntr)

####################################################################################################
# CREATE NLOED MODEL
####################################################################################################

# create list for observation structure 
ode_response = []
#create list to store response names
response_names=[]
# loop over samples (time points)
for i in range(len(sample_list)):
    #create a unique name for mrna and prot samples
    ecoli_name = 'EColi_'+'t'+"{0:0=2d}".format(times[i])
    salm_name = 'Salmonella_'+'t'+"{0:0=2d}".format(times[i])
    primary_name = 'PrimarySubstrate_'+'t'+"{0:0=2d}".format(times[i])
    secondary_name = 'SecondarySubstrate_'+'t'+"{0:0=2d}".format(times[i])
    #create mean and var tuple for mrna and prot observ.
    ecoli_stats = cs.vertcat(sample_list[i][0], 0.001)
    salm_stats = cs.vertcat(sample_list[i][1], 0.001)
    primary_stats = cs.vertcat(sample_list[i][2], 0.001)
    secondary_stats = cs.vertcat(sample_list[i][3], 0.001)
    #create casadi function for mrna and prot stats
    ecoli_func = cs.Function(ecoli_name,[u,p],[ecoli_stats])
    salm_func = cs.Function(salm_name,[u,p],[salm_stats])
    primary_func = cs.Function(primary_name,[u,p],[primary_stats])
    secondary_func = cs.Function(secondary_name,[u,p],[secondary_stats])
    #append the casadi function and distribution type to obs struct
    ode_response.append((ecoli_func,'Normal'))
    ode_response.append((salm_func,'Normal'))
    ode_response.append((primary_func,'Normal'))
    ode_response.append((secondary_func,'Normal'))
    #store response names for plotting
    response_names.append(ecoli_name)
    response_names.append(salm_name)
    response_names.append(primary_name)
    response_names.append(secondary_name)

#xnames = ['ecoli_ic','salm_ic','prim_ic','sec_ic']
xnames = ['ecoli_ic','salm_ic']
pnames = ['lambda_E','K_L','delta_E','lambda_S','K_G','alpha_E','alpha_G','alpha_S']

ode_model = Model(ode_response,xnames,pnames)

####################################################################################################
# GENERATE DESIGN
####################################################################################################

true_pars = np.log([0.5,0.01,0.1,0.4,0.01,0.2,0.1,0.1])

# continuous_inputs = {'Inputs':['ecoli_ic','salm_ic','prim_ic','sec_ic'],
#                      'Bounds':[(0.1,1),(0.1,0.2),(1,5),(0.01,0.1)],
#                      'Structure':[['ecoli_ic1','salm_ic1','prim_ic1','sec_ic1'],
#                                   ['ecoli_ic2','salm_ic2','prim_ic2','sec_ic2'],
#                                   ['ecoli_ic3','salm_ic3','prim_ic3','sec_ic3']]}
# continuous_inputs = {'Inputs':['ecoli_ic','salm_ic'],
#                      'Bounds':[(0.01,0.1),(0.1,0.1)],
#                      'Structure':[['ecoli_ic1','salm_ic1'],
#                                   ['ecoli_ic2','salm_ic2'],
#                                   ['ecoli_ic3','salm_ic3']]}
discrete_inputs = {'Inputs':['ecoli_ic','salm_ic'],
                     'Bounds':[(0.01,0.1),(0.01,0.1)]}
opt_design = Design(ode_model, true_pars, 'D', discrete_inputs=discrete_inputs)

sample_size = 30
exact_design = opt_design.round(sample_size)

print(exact_design)

####################################################################################################
# SAMPLE DESIGN AND FIT
####################################################################################################

#evaluate the proposed deign in terms of expected Covariance, Bias and MSE
opts = {'Covariance':True,'Bias':True,'MSE':True}
eval_dat = ode_model.evaluate(exact_design,true_pars,opts)
print(eval_dat)

#generate some data, to stand in for an initial experiment
data = ode_model.sample(exact_design,true_pars)
print(data)

#pass some additional options to fitting alglorithm, including Profile likelihood
fit_options={'Confidence':'Profiles',  #NOTE: ISSUE WITH PROFILES HERE
             'InitParamBounds':[(-5,1)]*8,
             'InitSearchNumber':3,
             'SearchBound':5.,
             'MaxSteps':100000}
#fit the model to the init data
fit_info = ode_model.fit(data, options=fit_options)
print(fit_info)

#extract fitted parameters
fit_params = fit_info['Estimate'].to_numpy().flatten()

####################################################################################################
# TESTING PREDICTIONS AND EVAL
####################################################################################################


# create data frame of inputs that need predictions
predict_inputs = pd.DataFrame({ 'ecoli_ic':[0.1]*len(response_names),
                                'salm_ic':[0.01]*len(response_names),
                                'prim_ic':[5]*len(response_names),
                                'sec_ic':[0.001]*len(response_names),
                                'Variable':response_names})

predict_design = cp.deepcopy(predict_inputs)
predict_design['Replicats'] = [5]*len(response_names)

#generate the predictions at fit parameters
predictions = ode_model.predict(predict_inputs,fit_params)
print(predictions)

#generate data at true parameters
plot_data = ode_model.sample(predict_design,true_pars)
print(plot_data)

# create regular rexpressions (re package) search strings
digit_re=re.compile('[a-z]+_t(\d+)')
type_re=re.compile('([a-z]+)_t\d+')

# use regular rexpressions (re package) to parse observations into times and rna/prot
plot_data['Time'] = plot_data['Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
plot_data['Type'] = plot_data['Variable'].apply(lambda x: type_re.search(x).group(1))
#organize the data for plotting
mrna_obs = plot_data.loc[plot_data['Type'] == 'mrna']
prot_obs = plot_data.loc[plot_data['Type'] == 'prot']
#plot true data
ax1 = mrna_obs.plot.scatter(x=('Time'),y=('Observation'),c='Blue')
ax2 = prot_obs.plot.scatter(x=('Time'),y=('Observation'),c='Blue')

# use regular rexpressions (re package) to parse predictions into times and rna/prot
predictions['Prediction','Time'] = predictions['Inputs','Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
predictions['Prediction','Type'] = predictions['Inputs','Variable'].apply(lambda x: type_re.search(x).group(1))
#organize the data for plotting
mrna_pred = predictions.loc[predictions['Prediction','Type'] == 'mrna']
prot_pred = predictions.loc[predictions['Prediction','Type'] == 'prot']
#plot the predicted mean values for the model at fit params
mrna_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red',ax=ax1)
prot_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red',ax=ax2)

plt.show()

t=0

####################################################################################################
# MANUALLY CREATE A DESIGN FOR TESTING
####################################################################################################

# design = pd.DataFrame({ 'mrna_ic':[1],
#                         'prot_ic':[1],
#                         'cntrl_1':[0.1],
#                         'cntrl_2':[1],
#                         'cntrl_3':[0.1]})

# design=design.reindex(design.index.repeat(len(response_names)))
# design['Variable'] = response_names
# design['Replicats'] = replicates
# design = design.sort_values(by='Variable').reset_index()

# fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# print(fit_pars)

# #evaluate goes here

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# cov_mat = np.diag([0.1,0.1,0.1,0.1])
# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)





#!/usr/bin/env python
# coding: utf-8

# # Evaluating Design Performance
# 
# In this example, we assume you have encoded an NLoed model and generated an optimal design. Before implementing the design you may wish to get a better and more interpretable understanding of its expected performance.
# 
# To begin we import the results for the *N2_designoptimization* notebook:

# In[ ]:


from scripts.N2_design_optimization import *




#from setuptools import setup
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nloed", 
    version="0.0.2",
    author='Nate Braniff',
    author_email='nbraniff@uwaterloo.ca',
    description="A package for nonlinear optimal experimental design",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NateBraniff/NLoed",
    keywords = ['Optimal','Experimental','Design'],
    packages=setuptools.find_packages(),
    python_requires='>=3.6.1',
    install_requires = [
        'casadi>=3.5.0',
        'numpy>=1.17.0',
        'pandas>=1.0.0',
        'scipy',
        'matplotlib',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    license = 'LGPLv3+'
    # py_modules=["model","design"],
    # package_dir={'':'nloed'},
)


""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
from nloed import Model
from nloed import Design

####################################################################################################
# SET UP MODEL
####################################################################################################

#define input and parameters
x = cs.SX.sym('x',1)
xnames = ['x']
p = cs.SX.sym('p',2)
pnames = ['Intercept','Slope']

#define the deterministic model
lin_predictor = p[0] + p[1]*x[0] 

#link the deterministic model to the sampling statistics (here normal mean and variance)
mean, var = lin_predictor, 0.1
normal_stats = cs.vertcat(mean, var)

#create a casadi function mapping input and parameters to sampling statistics (mean and var)
y = cs.Function('y',[x,p],[normal_stats])

# enter the function as a tuple with label indicating normal error, into observation list
observ_list = [(y,'Normal')]

#instantiate nloed model class
nloed_model = Model(observ_list,xnames,pnames)

####################################################################################################
# GENERATE OPTIMAL DESIGN
####################################################################################################

#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['x'],'Bounds':[(-1,1)],'Structure':[['x1'],['x2']]}
#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
#discrete_inputs={'Inputs':['x'],'Grid':[[-1.5,-0.1,0.1,1.5]]}

# generate the optimal approximate (relaxed) design
relaxed_design = Design(nloed_model,[1,1],'D',continuous_inputs=continuous_inputs)
# generate the optimal approximate (relaxed) design
#relaxed_design = Design(nloed_model,[1,1],'D',discrete_inputs=discrete_inputs)

sample_size = 10
#generate a rounded exact design 
exact_design = relaxed_design.round(sample_size)

print(exact_design)

####################################################################################################
# GENERATE SAMPLE DATA & FIT MODEL
####################################################################################################

#generate some data, to stand in for an initial experiment
data = nloed_model.sample(exact_design,[1,1])

print(data)

#pass some additional options to fitting alglorithm, including Profile likelihood
fit_options={'Confidence':'Profiles',
             'InitParamBounds':[(-5,5),(-5,5)],
             'InitSearchNumber':7}
#fit the model to the init data
fit_info = nloed_model.fit(data, options=fit_options)

print(fit_info)

fit_params = fit_info['Estimate'].to_numpy().flatten()

####################################################################################################
# EVALUATE DESIGN & PREDICT OUTPUT
####################################################################################################

#get estimated covariance, bias and MSE of parameter fit (use asymptotic method here) 
opts={'Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
diagnostic_info = nloed_model.evaluate(exact_design,fit_params,opts)
print(diagnostic_info)

#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({ 'x':np.random.uniform(-1,1,10),
                                'Variable':['y']*10})
cov_mat = diagnostic_info['Covariance'].to_numpy()
pred_options = {'PredictionInterval':True,
                'ObservationInterval':True,
                'Sensitivity':True}
predictions = nloed_model.predict(prediction_inputs,
                                  fit_params,
                                  covariance_matrix = cov_mat,
                                  options=pred_options)
print(predictions)

t=0
# design = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                         'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                         'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9,
#                         'Replicats':[5]*9*3})

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# #mixed model
# mean, var = lin_predictor, 0.1
# normal_stats = cs.vertcat(mean, var)
# y_norm_func = cs.Function('y_norm',[xs,ps],[normal_stats])

# rate = cs.exp(lin_predictor)
# poisson_stats = rate
# y_pois_func = cs.Function('y_pois',[xs,ps],[poisson_stats])

# prob = cs.exp(lin_predictor)/(1+cs.exp(lin_predictor))
# bern_stats = prob
# y_bern_func = cs.Function('y_bern',[xs,ps],[bern_stats])

# mixed_response = [  (y_norm_func,'Normal'),
#                     (y_bern_func,'Bernoulli'),
#                     (y_pois_func,'Poisson')]

# mixed_model = Model(mixed_response,xnames,pnames)

# opts={'Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
# eval_dat = mixed_model.evaluate(design,[0.5,1.1,2.1,0.3],opts)

# opts={'Method':'MonteCarlo','Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
# eval_dat = mixed_model.evaluate(design,[0.5,1.1,2.1,0.3],opts)

# mixed_data = mixed_model.sample(design,[0.5,1.1,2.1,0.3])
# fit_options={'Confidence':'Profiles',
#              'InitParamBounds':[(-5,5),(-5,5),(-5,5),(-5,5)],
#              'InitSearchNumber':7}
# mixed_fit = mixed_model.fit(mixed_data, options=fit_options)



# fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# print(fit_pars)

# #evaluate goes here

# cov_mat = np.diag([0.1,0.1,0.1,0.1])
# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)


# t=0

# x = cs.SX.sym('x')
# xnames = ['x']
# p = cs.SX.sym('p')
# pnames = ['p']

# design = pd.DataFrame({'x':[-2,-1,0,1,2],'y':[10,10,10,10,10]})

# # #normal model
# mean, var = x*p, x**2*p**2+.1
# normal_stats = cs.vertcat(mean, var)
# y_norm_func = cs.Function('y',[x,p],[normal_stats])
# normal_response = [(y_norm_func,'Normal')]
# normal_model = Model(normal_response,xnames,pnames)
# normal_data = normal_model.sample(design,[1])
# normal_fit = normal_model.fit(normal_data,[2])

# #poisson model
# rate = cs.exp(x*p)
# poisson_stats = rate
# y_pois_func = cs.Function('y',[x,p],[poisson_stats])
# poisson_response = [(y_pois_func,'Poisson')]
# poisson_model = Model(poisson_response,xnames,pnames)
# poisson_data = poisson_model.sample(design,[1])
# poisson_fit = poisson_model.fit(poisson_data,[2])

# #lognormal model
# geomean, geovar = x*p, cs.exp(x*p)
# logn_stats = cs.vertcat(geomean, geovar)
# y_logn_func = cs.Function('y',[x,p],[logn_stats])
# logn_response = [(y_logn_func,'Lognormal')]
# logn_model = Model(logn_response,xnames,pnames)
# logn_data = logn_model.sample(design,[1])
# logn_fit = logn_model.fit(logn_data,[2])

# #binomial model
# prob, num = cs.exp(x*p)/(1+cs.exp(x*p)), 10
# binom_stats = cs.vertcat(prob, num)
# y_binom_func = cs.Function('y',[x,p],[binom_stats])
# binom_response = [(y_binom_func,'Binomial')]
# binom_model = Model(binom_response,xnames,pnames)
# binom_data = binom_model.sample(design,[1])
# binom_fit = binom_model.fit(binom_data,[2])

# #bernoulli model
# prob = cs.exp(x*p)/(1+cs.exp(x*p))
# bern_stats = prob
# y_bern_func = cs.Function('y',[x,p],[bern_stats])
# bern_response = [(y_bern_func,'Bernoulli')]
# bern_model = Model(bern_response,xnames,pnames)
# bern_data = bern_model.sample(design,[1])
# bern_fit = bern_model.fit(bern_data,[2])

# #exponential model
# rate = cs.exp(x*p)
# exp_stats = prob
# y_pois_func = cs.Function('y',[x,p],[exp_stats])
# exp_response = [(y_pois_func,'Exponential')]
# exp_model = Model(exp_response,xnames,pnames)
# exp_data = exp_model.sample(design,[1])
# exp_fit = exp_model.fit(exp_data,[2])

# t=0
# ##########################################################################################

# xs = cs.SX.sym('xs',2)
# xnames = ['x1','x2']
# ps = cs.SX.sym('ps',4)
# pnames = ['Intercept','x1-Main','x2-Main','Interaction']

# lin_predictor = ps[0] + ps[1]*xs[0] + ps[2]*xs[1] + ps[3]*xs[0]*xs[1]

# design = pd.DataFrame({'x1':[-1,-1,-1,0,0,0,1,1,1],'x2':[-1,0,1,-1,0,1,-1,0,1],'y':[5,5,5,5,5,5,5,5,5]})

# #normal model
# mean, var = lin_predictor, lin_predictor**2 + 0.1
# normal_stats = cs.vertcat(mean, var)
# y_norm_func = cs.Function('y',[xs,ps],[normal_stats])
# normal_response = [(y_norm_func,'Normal')]
# normal_model = Model(normal_response,xnames,pnames)
# normal_data = normal_model.sample(design,[0.5,1.1,2.1,0.3])
# normal_fit = normal_model.fit(normal_data,[1,1,1,1])

# #poisson model
# rate = cs.exp(lin_predictor)
# poisson_stats = rate
# y_pois_func = cs.Function('y',[xs,ps],[poisson_stats])
# poisson_response = [(y_pois_func,'Poisson')]
# poisson_model = Model(poisson_response,xnames,pnames)
# poisson_data = poisson_model.sample(design,[0.5,1.1,2.1,0.3])
# poisson_fit = poisson_model.fit(poisson_data,[1,1,1,1])

# #lognormal model
# geomean, geovar = lin_predictor, cs.exp(lin_predictor)
# logn_stats = cs.vertcat(geomean, geovar)
# y_logn_func = cs.Function('y',[xs,ps],[logn_stats])
# logn_response = [(y_logn_func,'Lognormal')]
# logn_model = Model(logn_response,xnames,pnames)
# logn_data = logn_model.sample(design,[0.5,1.1,2.1,0.3])
# logn_fit = logn_model.fit(logn_data,[1,1,1,1])

# #binomial model
# prob, num = cs.exp(lin_predictor)/(1+cs.exp(lin_predictor)), 10
# binom_stats = cs.vertcat(prob, num)
# y_binom_func = cs.Function('y',[xs,ps],[binom_stats])
# binom_response = [(y_binom_func,'Binomial')]
# binom_model = Model(binom_response,xnames,pnames)
# binom_data = binom_model.sample(design,[0.5,1.1,2.1,0.3])
# binom_fit = binom_model.fit(binom_data,[1,1,1,1])

# #bernoulli model
# prob = cs.exp(lin_predictor)/(1+cs.exp(lin_predictor))
# bern_stats = prob
# y_bern_func = cs.Function('y',[xs,ps],[bern_stats])
# bern_response = [(y_bern_func,'Bernoulli')]
# bern_model = Model(bern_response,xnames,pnames)
# bern_data = bern_model.sample(design,[0.5,1.1,2.1,0.3])
# bern_fit = bern_model.fit(bern_data,[1,1,1,1])

# #exponential model
# rate = cs.exp(lin_predictor)
# exp_stats = prob
# y_pois_func = cs.Function('y',[xs,ps],[exp_stats])
# exp_response = [(y_pois_func,'Exponential')]
# exp_model = Model(exp_response,xnames,pnames)
# exp_data = exp_model.sample(design,[0.5,1.1,2.1,0.3])
# exp_fit = exp_model.fit(exp_data,[1,1,1,1])


# t=0

# xs = cs.SX.sym('xs',2)
# xnames = ['x1','x2']
# ps = cs.SX.sym('ps',3)
# pnames = ['Intercept','x1-Main','x2-Main']

# lin_predictor = ps[0] + ps[1]*xs[0] + ps[2]*xs[1]

# design = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1],
#                         'x2':[-1,0,1,-1,0,1,-1,0,1],
#                         'y_norm':[50,50,50,50,50,50,50,50,50]})

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1],
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1],
#                                 'Variable':['y_norm']*9})

# #mixed model
# mean, var = lin_predictor, lin_predictor**2+0.1
# normal_stats = cs.vertcat(mean, var)
# y_norm_func = cs.Function('y_norm',[xs,ps],[normal_stats])

# lin_response = [ (y_norm_func,'Normal')]

# lin_model = Model(lin_response,xnames,pnames)

# lin_data = lin_model.sample(design,[-2,0.5,-1.1])
# options={'Confidence':'Contours','InitParameters':[-1,0,-1]}
# lin_fit = lin_model.fit(lin_data,options)


# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True,
#                 'PredictionSampleNumber':10000}
# cov_mat = np.diag([0.1,0.1,0.1])
# predictions_mc = lin_model.predict(predict_inputs,[-2,0.5,-1.1],covariance_matrix= cov_mat,options=pred_options)

# pred_options = {'Method':'Delta',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# cov_mat = np.diag([0.1,0.1,0.1])
# predictions_dlta = lin_model.predict(predict_inputs,[-2,0.5,-1.1],covariance_matrix = cov_mat,options=pred_options)


# options={'Confidence':'Profiles','InitParamBounds':[(-5,5),(-5,5),(-5,5)],'InitSearchNumber':7}

# lin_data = lin_model.sample(design,[-2,0.5,-1.1])
# lin_fit = lin_model.fit(lin_data,options=options)
# print(lin_fit)
# fit_param_vec=lin_fit['Estimate'].to_numpy()
# pred_options={'Method':'Exact','MeanInterval':False,'PredictionInterval':True}
# cov_mat=np.diag([0.1,0.1,0.1])
# predictions=lin_model.predict(predict_inputs,fit_param_vec,covariance_matrix= cov_mat,options=pred_options)

# lin_data = lin_model.sample(design,[0.2,1.9,3.25])
# lin_fit = lin_model.fit(lin_data,options=options)
# print(lin_fit)

# lin_data = lin_model.sample(design,[-0.5,-3.3,0.2])
# lin_fit = lin_model.fit(lin_data,options=options)
# print(lin_fit)

# lin_data = lin_model.sample(design,[4.3,0,-0.4])
# lin_fit = lin_model.fit(lin_data,options=options)
# print(lin_fit)

# xs = cs.SX.sym('xs',2)
# xnames = ['x1','x2']
# ps = cs.SX.sym('ps',4)
# pnames = ['Intercept','x1-Main','x2-Main','Interaction']

# lin_predictor = ps[0] + ps[1]*xs[0] + ps[2]*xs[1] + ps[3]*xs[0]*xs[1]

# design = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1],
#                         'x2':[-1,0,1,-1,0,1,-1,0,1],
#                         'y_norm':[5,5,5,5,5,5,5,5,5],
#                         'y_bern':[5,5,5,5,5,5,5,5,5],
#                         'y_pois':[5,5,5,5,5,5,5,5,5]})

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# #mixed model
# mean, var = lin_predictor, 0.1
# normal_stats = cs.vertcat(mean, var)
# y_norm_func = cs.Function('y_norm',[xs,ps],[normal_stats])

# rate = cs.exp(lin_predictor)
# poisson_stats = rate
# y_pois_func = cs.Function('y_pois',[xs,ps],[poisson_stats])

# prob = cs.exp(lin_predictor)/(1+cs.exp(lin_predictor))
# bern_stats = prob
# y_bern_func = cs.Function('y_bern',[xs,ps],[bern_stats])

# mixed_response = [  (y_norm_func,'Normal'),
#                     (y_bern_func,'Bernoulli'),
#                     (y_pois_func,'Poisson')]

# mixed_model = Model(mixed_response,xnames,pnames)
# mixed_data = mixed_model.sample(design,[0.5,1.1,2.1,0.3])
# fit_options={'Confidence':'Profiles',
#             'InitParamBounds':[(-5,5),(-5,5),(-5,5),(-5,5)],
#             'InitSearchNumber':7}
# mixed_fit = mixed_model.fit(mixed_data, options=fit_options)

# fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# print(fit_pars)

# #evaluate goes here

# cov_mat = np.diag([0.1,0.1,0.1,0.1])
# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)


# t=0
# stats2=cs.vertcat(mean,var)
# stats3=cs.vertcat(mean,10)

# f1 = cs.Function('f1',[x,p],[stats1])
# f2 = cs.Function('f2',[x,p],[stats2])
# f3 = cs.Function('f2',[x,p],[stats3])


# resBern = [('y','Bernoulli',f1)]
# resBin = [('y','Binomial',f3)]
# resLogn = [('y','Lognormal',f2)]
# resPois = [('y','Poisson',f1)]
# resExp = [('y','Exponential',f1)]

# norm_mod = Model(resNorm,xnames,pnames)
# logn_mod = Model(resLogn,xnames,pnames)
# bern_mod = Model(resBern,xnames,pnames)
# bin_mod = Model(resBin,xnames,pnames)
# pois_mod = Model(resPois,xnames,pnames)
# exp_mod = Model(resExp,xnames,pnames)

# design = pd.DataFrame({'x':[1,20],'y':[5,5]})

# norm_data = norm_mod.sample(design,[1])
# logn_data = logn_mod.sample(design,[1])
# bern_data = bern_mod.sample(design,[.03])
# bin_data = bin_mod.sample(design,[.03])
# pois_data = pois_mod.sample(design,[1])
# exp_data = exp_mod.sample(design,[1])

# t=0


# #Declare model inputs and parametrs
# inputs = cs.SX.sym('inputs',1)
# params = cs.SX.sym('params',2)
# #define sampling stat model for normal response
# mean = p[0] + x*p[1]
# variance = cs.SX(0.01)
# #zip the sampling stats into a vector and create a casadi function
# stats = cs.vertcat(mean, variance)
# func = cs.Function('func',[params,inputs],[stats])
# #Enter the above model into the list of reponse variables
# response= [('y1','Normal',func)]
# #give the input
# input_names = ['x1']
# param_names = ['p1','p2']

# #Instantiate class
# linear1d=Model(response,input_names,param_names)

# design1 = pd.DataFrame({'x1':[0,1,2],'y1':[5,1,5]})
# design2 = pd.DataFrame({'x1':[-1,0,1],'y1':[4,3,4]})

# dataset_list = linear1d.sample([design1,design2],[0,1],5)

# pars_info = linear1d.fit(dataset_list,[0,1],options={'Confidence':'Contours'})

# prediction_inputs = dataset_list[0][0][input_names+['Variable']]
# par_est = pars_info['Estimate'][param_names].to_numpy()

# #pred_structA=linear1d.eval_model(inputs, par_est, [[1,0],[0,1]], True,options={'ErrorMethod':'Delta'})
# #pred_structB=linear1d.eval_model(inputs, par_est, [[1,0],[0,1]], True,options={'ErrorMethod':'MonteCarlo','SampleSize':10000})
# # print(pred_structA['y1']['Mean']['Bounds'])
# # print(pred_structB['y1']['Mean']['Bounds'])

# # pred_struct4=linear1d.eval_model([1,1],[[i] for i in range(10)],[[1,0],[0,1]],True,options={'ErrorMethod':'Delta'})
# # pred_struct5=linear1d.eval_model([1,1],[[i] for i in range(10)],[[1,0],[0,1]],True,options={'ErrorMethod':'MonteCarlo'})
# # pred_struct1=linear1d.eval_model([1,1],1)
# # pred_struct2=linear1d.eval_model([1,1],1,param_covariance=[[1,0],[0,1]])
# # pred_struct3=linear1d.eval_model([1,1],1,sensitivity=True)



# # paramCov1=np.cov(pars,rowvar=False)
# # paramMean1=np.mean(pars,axis=0)

# # dataset2=linear1d.sample([Experiment,Experiment],[0,1])

# modelinfo={'Model':linear1d, 'Parameters': [0.5, 1],'Objective':'D'}
# approx={'Inputs':['x1'],'Bounds':[(-1,1)]}
# obs={'Observations':[['y1']]}

# opt_approx=design([modelinfo],approxinputs=approx,observgroups=obs)
# print(opt_approx)

# opt_approx=design([modelinfo],approx)
# print(opt_approx)

# modelinfo={'Model':linear1d, 'Parameters': [0.5, 1],'Objective':'D'}
# struct=[['x1_lvl1'],['x1_lvl2'],['x1_lvl3']]
# exact={'Inputs':['x1'],'Bounds':[(-1,1)],'Structure':struct}
# obs={'Observations':[['y1']]}

# opt_exact=design([modelinfo],exactinputs=exact,observgroups=obs)
# print(opt_exact)

# opt_exact=design([modelinfo],exactinputs=exact)
# print(opt_exact)

# print('Done')




""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
from nloed import Design

###block 1
from nloed import Model
#create casadi symbols for the inputs
x = cs.SX.sym('x',2)
#create casadi symbols for the parameters
theta = cs.SX.sym('theta',4)
#define y1 sampling statistics; mean and variance
mean_y1 = theta[0] + theta[1]*x[0] + theta[3]*x[0]*x[1] 
var_y1 = 0.1
#define y2 sampling statistics; mean and variance
rate_y2 = cs.exp(theta[0] + theta[2]*x[1] + theta[3]*x[0]*x[1])
#create a casadi function for y1 stats
eta_y1 = cs.vertcat(mean_y1, var_y1)
func_y1 = cs.Function('y1',[x,theta],[eta_y1])
#create a casadi function for y2 stats
eta_y2 = rate_y2
func_y2 = cs.Function('y2',[x,theta],[eta_y2])

###block 2
#create observation list
observ_list = [(func_y1,'Normal'),(func_y2,'Poisson')]
#creat input name list
input_names = ['x1','x2']
#create parameter name list
parameter_names = ['Theta0','Theta1','Theta2','Theta3']
#create NLOED Model
model_object = Model(observ_list, input_names, parameter_names)

###block 3
design = pd.DataFrame({ 'x1':[0,-1,2,3]*2,
                        'x2':[1,1,-1,0]*2,
                        'Variable':['y1']*4 + ['y2']*4,
                        'Replicates':[3,1,2,2]*2})
print('')
print('')
print(design)
print('')
print('')

###block 4
design = pd.DataFrame({ 'x1':[0,-1,2,3]*2,
                        'x2':[1,1,-1,0]*2,
                        'Variable':['y1']*4 + ['y2']*4,
                        'Replicates':[3,1,2,2]*2})
#set nominal parameter values
param = [0.1, 2, 0.4, 1.3]
#declare specific options
eval_opts={'Method':'MonteCarlo',
           'FIM':True,
           'Covariance':True,
           'Bias':True,
           'MSE':True,
           'SampleNumber':100}
#call the evaluate() function
eval_info = model_object.evaluate(design, param, eval_opts)
#print the resulting evaluation
#print(eval_info)
print('')
print('')
print(eval_info)
print('')
print('')

### block 5
#define a design
design = pd.DataFrame({ 'x1':[0,-1,2,3]*2,
                        'x2':[1,1,-1,0]*2,
                        'Variable':['y1']*4 + ['y2']*4,
                        'Replicates':[3,1,2,2]*2})
#set nominal parameter values
param = [0.1, 2, 0.4, 1.3]
#call the sample() function
dataset = model_object.sample(design, param, design_replicates=1)
#print the resulting dataset
#print(dataset)
print('')
print('')
print(dataset)
print('')
print('')

###block 6
# set specific fitting options
fit_opts={'Confidence':'None',
             'InitParamBounds':[(-1,1),(-1,1),(-1,1),(-1,1)],
             'InitSearchNumber':7,
             'RadialNumber':60}
#call the fit() function
fit_info = model_object.fit(dataset, options=fit_opts)
#print the fitting information
#print(fit_info)
print('')
print('')
print(fit_info)
print('')
print('')

### block 7
#create three datasets
data1, data2, data3 = dataset = model_object.sample(design, param, design_replicates=3)
# set up specific options for fitting
fit_opts={'Confidence':'Intervals'}
#combine datasets into a single list
datasets = [data1, data2, data3]
#call the fitting procedure
fit_info = model_object.fit(datasets, options=fit_opts)
#print the fitting information
print(fit_info)
print('')
print('')
print(fit_info)
print('')
print('')

#define the inputs for predict()
predict_inputs = pd.DataFrame({ 'x1':[-1,1,-1,1]*2,
                                'x2':[-1,-1,1,1]*2,
                               'Variable':['y1']*4 + ['y2']*4})
#specify the parameter values
params = [0.1, 2, 0.4, 1.3]
#generate the predictions
predictions = model_object.predict(predict_inputs, params)
#display the predicted outputs
print(predictions)
print('')
print('')
print(predict_inputs)
print('')
print('')
print(predictions)
print('')
print('')


#define a covariance matrix for the parameters
cov_mat = np.diag(np.array(param)*0.05)
#specify desired option values
predict_opts = {'Method':'MonteCarlo',
                'PredictionInterval':True,
                'ObservationInterval':True,
                'Sensitivity':True}
#generate the predictions
predictions = model_object.predict(predict_inputs, params, cov_mat, predict_opts)
#display the predicted outputs
print(predictions)
print('')
print('')
print(predictions)
print('')
print('')

####################################################################################################
# GENERATE DESIGN
####################################################################################################

true_param = [1,1,1,1]

#discrete_inputs={'Inputs':['x1'],'Candidates':[[-1],[-.5],[0],[.5],[1]]}
#continuous_inputs={'Inputs':['x2'],'Bounds':[(-1,1)],'Structure':[['level1'],['level2']]}
discrete_inputs={'Inputs':['x1','x2'],'Candidates':[[-1, 0, 1],[-1, 0, 1]]}
#discrete_inputs={'Inputs':['x1','x2'],'Grid':[[-1,-1],[0,-1],[1,-1],[-1,0],[0,0],[1,0],[-1,1],[0,1],[1,1],[3,3]]}
## continuous_inputs={'Inputs':['x1','x2'],'Bounds':[(-1,1),(-1,1)],'Structure':[['x1_lvl1','x2_lvl1'],['x1_lvl1','x2_lvl2'],['x1_lvl2','x2_lvl2']]}
design_object = Design(lin_model, true_param, 'D', discrete_inputs)


relaxed_design = design_object.relaxed()
print('')
print('')
print(relaxed_design)
print('')
print('')


sample_size = 10
exact_design = design_object.round(sample_size)

print('')
print('')
print(exact_design)
print('')
print('')

####################################################################################################
# GENERATE SAMPLE DATA & FIT MODEL
####################################################################################################

#generate some data, to stand in for an initial experiment
data = lin_model.sample(exact_design,true_param, design_replicats=3)

print('')
print('')
print(data)
print('')
print('')

#pass some additional options to fitting alglorithm, including Profile likelihood
fit_options={'Confidence':'Intervals',
             'InitParamBounds':[(-1,1),(-1,1),(-1,1),(-1,1)],
             'InitSearchNumber':7,
             'SearchBound':5.}
#fit the model to the init data
fit_info = lin_model.fit(data, options=fit_options)

print('')
print('')
print(fit_info)
print('')
print('')

fit_params = fit_info['Estimate'].to_numpy()[0].flatten()

fit_info = fit_params

print('')
print('')
print(fit_info)
print('')
print('')


design = pd.DataFrame({'x1':[-1,1,-1,1]*2,
                       'x2':[-1,-1,1,1]*2,
                       'Variable':['y1']*4 + ['y2']*4,
                       'Replicats':[3]*8})
print('')
print('')
print(design)
print('')
print('')


####################################################################################################
# EVALUATE DESIGN & PREDICT OUTPUT
####################################################################################################

#get estimated covariance, bias and MSE of parameter fit (use asymptotic method here) 
opts={'Method':'MonteCarlo','Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
diagnostic_info = lin_model.evaluate(exact_design,fit_params,opts)

print('')
print('')
print(diagnostic_info)
print('')
print('')

#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({ 'x1':np.random.uniform(0,5,10),
                                   'x2':np.random.uniform(0,5,10),
                                   'Variable':['y']*10})
cov_mat = diagnostic_info['Covariance'].to_numpy()
pred_options = {'Method':'MonteCarlo',
                'PredictionInterval':True,
                'ObservationInterval':True,
                'Sensitivity':True}
predictions = lin_model.predict(prediction_inputs,
                                  fit_params,
                                  covariance_matrix = cov_mat,
                                  options=pred_options)
print(predictions)

t=0


input_info = {'x1':{'Type':'Continuous','Bounds':(-1,1)}
                }

# design = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                         'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                         'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9,
#                         'Replicats':[5]*9*3})

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# #mixed model


# rate = cs.exp(lin_predictor)
# poisson_stats = rate
# y_pois_func = cs.Function('y_pois',[xs,ps],[poisson_stats])

# prob = cs.exp(lin_predictor)/(1+cs.exp(lin_predictor))
# bern_stats = prob
# y_bern_func = cs.Function('y_bern',[xs,ps],[bern_stats])

# mixed_response = [  (y_norm_func,'Normal'),
#                     (y_bern_func,'Bernoulli'),
#                     (y_pois_func,'Poisson')]

# mixed_model = Model(mixed_response,xnames,pnames)

# opts={'Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
# eval_dat = mixed_model.evaluate(design,[0.5,1.1,2.1,0.3],opts)

# opts={'Method':'MonteCarlo','Covariance':True,'Bias':True,'MSE':True,'SampleNumber':100}
# eval_dat = mixed_model.evaluate(design,[0.5,1.1,2.1,0.3],opts)

# mixed_data = mixed_model.sample(design,[0.5,1.1,2.1,0.3])
# fit_options={'Confidence':'Profiles',
#             'InitParamBounds':[(-5,5),(-5,5),(-5,5),(-5,5)],
#             'InitSearchNumber':7}
# mixed_fit = mixed_model.fit(mixed_data, options=fit_options)



# fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# print(fit_pars)

# #evaluate goes here

# cov_mat = np.diag([0.1,0.1,0.1,0.1])
# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)



#Declare model inputs
# x=cs.SX.sym('x',4)

# #Declare the parameters-of-interest
# beta=cs.SX.sym('beta',6)

# #Define a model t
# # theta1 = cs.Function('theta1',[beta,x],[beta[0] + x[0]*beta[1] + x[1]*beta[2],1])
# # theta2 = cs.Function('theta2',[beta,x],[beta[3] + x[0]*beta[4] + x[1]**2*beta[5],1])
# theta1 = cs.Function('theta1',[beta,x],[beta[0] + x[0]*beta[2] + x[1]*x[3]*beta[4],1])
# theta2 = cs.Function('theta2',[beta,x],[beta[1] + x[1]*beta[3] + x[0]*x[2]*beta[5],1])

# #Enter the above model into the list of reponse variables
# response= [('y1','Normal',theta1),('y2','Normal',theta2)]
# #xnames=['x1','x2']
# xnames=['x1','x2','x3','x4']
# #betanames=['b1','b2','b3','b4']
# #betanames=['b1','b2','b3']
# betanames=['b1','b2','b3','b4','b5','b6']

# #Instantiate class
# linear2dx2d=model(response,xnames,betanames)

# model1={'Model':linear2dx2d, 'Parameters': [0.5,1,2,0.5,1,2],'Objective':'D'}
# #model2={'Model':linear2dx2d, 'Parameters': [2, 2, 2, 2],'Objective':'D'}

# #models=[model1,model2]
# models=[model1]

# approx={'Inputs':['x1','x2','x3','x4'],'Bounds':[(-1,1),(-1,1),(-1,1),(-1,1)]}
# obs={'Observations':[['y1','y2']]}


# design1=design(models,approx,observgroups=obs)
# print(design1)

# design1=design(models,approx)
# print(design1)


#NOTE: Add to model later: Weight, Prior (cov, norm only, user can transform var), POI (parameters-of-interest)
#NOTE: Add to approx later: Constraints, Resolution (stepsize), Grid (user defined set of possible input levels, or full grid?)
#NOTE: Add to exact later: Replicates (code to default with full reps of all unique vec), Constraints, Start, 
#NOTE: Add to obs later: Weights, if not passed, treat each observation as individual option


# #Instantiate class
# linear2dx2d=model(response,xnames,betanames)

# model1={'Model':linear2dx2d, 'Parameters': [0.5,1,2],'Objective':'D'}
# #model2={'Model':linear2dx2d, 'Parameters': [2, 2, 2, 2],'Objective':'D'}
# #NOTE: Add to model later: Weight, Prior (cov, norm only, user can transform var), POI (parameters-of-interest)
# #models=[model1,model2]
# models=[model1]

# approx={'Inputs':['x3','x4'],'Bounds':[(-1,1),(-1,1)]}
# #NOTE: Add to approx later: Constraints, Resolution (stepsize), Grid (user defined set of possible input levels, or full grid?)
# struct=[['x1_lvl1', 'x2_lvl1'],['x1_lvl1', 'x2_lvl2'],['x1_lvl2', 'x2_lvl1'],['x1_lvl2', 'x2_lvl2']]
# exact={'Inputs':['x1','x2'],'Bounds':[(-1,1),(-1,1)],'Structure':struct}
# #NOTE: Add to exact later: Replicates (code to default with full reps of all unique vec), Constraints, Start, 
# obs={'Observations':[['y1','y2']]}
# #NOTE: Add to obs later: Weights, if not passed, treat each observation as individual option

# design1=design(models,approx,exact,obs)
# print(design1)

# design1=design(models,approx,exact)
# print(design1)




import casadi as cs
import pandas as pd
import numpy as np
import re as re
import copy as cp
import matplotlib.pyplot as plt
from nloed import Model
from nloed import design

#states and controls
y = cs.MX.sym('y',2)
u = cs.MX.sym('u')
p = cs.MX.sym('p',4)
rhs = cs.vertcat(cs.exp(p[0])*u -cs.exp(p[1])*y[0], cs.exp(p[2])*y[0]-cs.exp(p[3])*y[1])
#ode = cs.Function('ode',[y,u,p],[rhs])

##########################################################################################
steps_per_sample = [1,2,3,5] 
#steps_per_sample = [1,9] 
samples_per_cntrl = 1
cntrls_per_run = 3

y0 = cs.MX.sym('y0',2)
uvec = cs.MX.sym('uvec',3)

x = cs.vertcat(y0,uvec)

dt=1
# # Create symbolics for RK4 integration, as shown in Casadi examples
ode_sys = {'x':y, 'p':cs.vertcat(u,p), 'ode':rhs}

#Loop over the ste function to create symbolics for integrating across a sample interval
y_sample = y0
sample_list=[]
times=[]
time_cntr=0
for i in range(cntrls_per_run):
  for num_stps in steps_per_sample:
    stepper = cs.integrator('F_'+str(i)+'_'+str(num_stps), 'cvodes', ode_sys, {'tf':dt*num_stps})
    y_sample = stepper(x0=y_sample, p=cs.vertcat(uvec[i], p))['xf']
    sample_list.append(y_sample)
    time_cntr+=num_stps*dt
    times.append(time_cntr)

design = pd.DataFrame({ 'mrna_ic':[1],
                        'prot_ic':[1],
                        'cntrl_1':[0.1],
                        'cntrl_2':[1],
                        'cntrl_3':[0.1]})

ode_response = []
response_names=[]
replicates=[]
for i in range(len(sample_list)):

  mrna_name = 'mrna_'+'t'+"{0:0=2d}".format(times[i])
  mrna_stats = cs.vertcat(sample_list[i][0], 0.001)
  mrna_func = cs.Function(mrna_name,[x,p],[mrna_stats])
  ode_response.append((mrna_func,'Normal'))
  response_names.append(mrna_name)
  replicates.append(5)

  prot_name = 'prot_'+'t'+"{0:0=2d}".format(times[i])
  prot_stats = cs.vertcat(sample_list[i][1], 0.001)
  prot_func = cs.Function(prot_name,[x,p],[prot_stats])
  ode_response.append((prot_func,'Normal'))
  response_names.append(prot_name)
  replicates.append(5)

design=design.reindex(design.index.repeat(len(response_names)))
design['Variable'] = response_names
design['Replicats'] = replicates
design = design.sort_values(by='Variable').reset_index()

xnames = ['mrna_ic','prot_ic','cntrl_1','cntrl_2','cntrl_3']
pnames = ['alpha','delta','beta','gamma']

ode_model = Model(ode_response,xnames,pnames,{'ScalarSymbolics':False})

predict_inputs = pd.DataFrame({ 'mrna_ic':[1]*len(response_names),
                                'prot_ic':[1]*len(response_names),
                                'cntrl_1':[0.1]*len(response_names),
                                'cntrl_2':[1.0]*len(response_names),
                                'cntrl_3':[0.1]*len(response_names),
                                'Variable':response_names})

true_pars = [np.log(0.5),np.log(1.1),np.log(2.1),np.log(0.3)]
predictions = ode_model.predict(predict_inputs,true_pars)

digit_re=re.compile('[a-z]+_t(\d+)')
type_re=re.compile('([a-z]+)_t\d+')

predictions['Prediction','Time'] = predictions['Inputs','Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
predictions['Prediction','Type'] = predictions['Inputs','Variable'].apply(lambda x: type_re.search(x).group(1))

mrna_pred = predictions.loc[predictions['Prediction','Type'] == 'mrna']
prot_pred = predictions.loc[predictions['Prediction','Type'] == 'prot']

ax1=mrna_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red')
ax2=prot_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red')

ode_data = ode_model.sample(design,true_pars,design_replicats=3)

data=cp.deepcopy(ode_data[0])

data['Time'] = data['Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
data['Type'] = data['Variable'].apply(lambda x: type_re.search(x).group(1))

mrna_obs = data.loc[data['Type'] == 'mrna']
prot_obs = data.loc[data['Type'] == 'prot']

mrna_obs.plot.scatter(x=('Time'),y=('Observation'),ax=ax1)
prot_obs.plot.scatter(x=('Time'),y=('Observation'),ax=ax2)
plt.show()


fit_options={'Confidence':'Intervals',
            'InitParamBounds':[(-3,2)]*4,
            'InitSearchNumber':3,
            'MaxSteps':100000}
ode_fit = ode_model.fit(ode_data, options=fit_options)

t=0

# fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# print(fit_pars)

# #evaluate goes here

# predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
#                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
#                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# cov_mat = np.diag([0.1,0.1,0.1,0.1])
# pred_options = {'Method':'MonteCarlo',
#                 'PredictionInterval':True,
#                 'ObservationInterval':True,
#                 'Sensitivity':True}
# predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)


# """ 
# Add a docstring
# """
# import numpy as np
# import pandas as pd
# import casadi as cs
# from nloed import Model
# from nloed import Design

# import re as re
# import copy as cp
# import matplotlib.pyplot as plt


# ####################################################################################################
# # SET UP RHS OF ODE
# ####################################################################################################

# #states and controls
# # y = cs.MX.sym('y',2)
# # u = cs.MX.sym('u')
# # p = cs.MX.sym('p',4)
# #####  -or-  #####
# y = cs.SX.sym('y',2)
# u = cs.SX.sym('u')
# p = cs.SX.sym('p',4)

# rhs = cs.vertcat(cs.exp(p[0])*u -cs.exp(p[1])*y[0], cs.exp(p[2])*y[0]-cs.exp(p[3])*y[1])
# ode = cs.Function('ode',[y,u,p],[rhs])
# dt = 1

# ####################################################################################################
# # DEFINE RK-4 STEP ALGEBRA
# ####################################################################################################

# # Create symbolics for RK4 integration, as shown in Casadi examples
# k1 = ode(y, u, p)
# k2 = ode(y + dt/2.0*k1, u, p)
# k3 = ode(y + dt/2.0*k2, u, p)
# k4 = ode(y + dt*k3, u, p)
# y_step = y+dt/6.0*(k1+2*k2+2*k3+k4)
# # Create a function to perform one step of the RK integration
# step = cs.Function('step',[y, u, p],[y_step])
# #####  -or-  #####
# # ode_sys = {'x':y, 'p':cs.vertcat(u,p), 'ode':rhs}
# # opts = {'tf':dt}
# # step = cs.integrator('F', 'cvodes', ode_sys, opts)

# ####################################################################################################
# # DEFINE TIME DISCRETIZATION, AND CONSTRUCT DETERMINISTIC MODEL VIA INTEGRATION
# ####################################################################################################

# steps_per_sample = [1,2,3,5] 
# #steps_per_sample = [1,9] 
# samples_per_cntrl = 1
# cntrls_per_run = 3

# # y0 = cs.MX.sym('y0',2)
# # uvec = cs.MX.sym('uvec',3)
# #####  -or-  #####
# y0 = cs.SX.sym('y0',2)
# uvec = cs.SX.sym('uvec',3)
# x = cs.vertcat(y0,uvec)

# #Loop over the ste function to create symbolics for integrating across a sample interval
# y_sample = y0
# sample_list=[]
# times=[]
# cntr=0
# for i in range(cntrls_per_run):
#   for num_stps in steps_per_sample:
#     for k in range(num_stps):
#       y_sample = step(y_sample, uvec[i], p)
#       #####  -or-  #####
#       #y_sample = step(x0=y_sample, p=cs.vertcat(uvec[i], p))['xf']
#       cntr+=1
#     sample_list.append(y_sample)
#     times.append(cntr*dt)

# ####################################################################################################
# # CONSTRUCT MRNA and PROTEIN MEAN/VAR STATS, CREATE INTO MODEL
# ####################################################################################################

# ode_response = []
# response_names=[]
# replicates=[]
# for i in range(len(sample_list)):
#   reps=10
#   mrna_name = 'mrna_'+'t'+"{0:0=2d}".format(times[i])
#   mrna_stats = cs.vertcat(sample_list[i][0], 0.001)
#   mrna_func = cs.Function(mrna_name,[x,p],[mrna_stats])
#   ode_response.append((mrna_func,'Normal'))
#   response_names.append(mrna_name)
#   replicates.append(reps)

#   prot_name = 'prot_'+'t'+"{0:0=2d}".format(times[i])
#   prot_stats = cs.vertcat(sample_list[i][1], 0.001)
#   prot_func = cs.Function(prot_name,[x,p],[prot_stats])
#   ode_response.append((prot_func,'Normal'))
#   response_names.append(prot_name)
#   replicates.append(reps)

# xnames = ['mrna_ic','prot_ic','cntrl_1','cntrl_2','cntrl_3']
# pnames = ['alpha','delta','beta','gamma']

# ode_model = Model(ode_response,xnames,pnames)

# ####################################################################################################
# # GENERATE DESIGN
# ####################################################################################################

# true_pars = np.log([0.5,1.1,2.1,0.3])

# approx_inputs = {'Inputs':['mrna_ic','prot_ic','cntrl_1','cntrl_2','cntrl_3'],
#                  'Bounds':[(0,1),(0,1),(0,1),(0,1),(0,1)]}
# opt_design = Design(ode_model,true_pars,'D',approx_inputs)
# #exact_inputs = {'Inputs':['mrna_ic','prot_ic','cntrl_1','cntrl_2','cntrl_3'],
# #                 'Bounds':[(0,1),(0,1),(0,1),(0,1),(0,1),(0,1)],
# #                 'Structure':[['mrna_ic1','prot_ic1','c1_lvl1','c2_lvl1','c3_lvl1'],
# #                              ['mrna_ic2','prot_ic2','c1_lvl2','c2_lvl2','c3_lvl2'],
# #                              ['mrna_ic3','prot_ic3','c1_lvl3','c2_lvl3','c3_lvl3']]}
# #opt_design = Design(ode_model,true_pars,'D',exact_inputs=exact_inputs)

# sample_size = 30
# exact_design = opt_design.round(sample_size)

# print(exact_design)


# ####################################################################################################
# # MANUALLY CREATE A DESIGN FOR TESTING
# ####################################################################################################

# design = pd.DataFrame({ 'mrna_ic':[1],
#                         'prot_ic':[1],
#                         'cntrl_1':[0.1],
#                         'cntrl_2':[1],
#                         'cntrl_3':[0.1]})

# design=design.reindex(design.index.repeat(len(response_names)))
# design['Variable'] = response_names
# design['Replicats'] = replicates
# design = design.sort_values(by='Variable').reset_index()

# ####################################################################################################
# # TESTING PREDICTIONS AND EVAL
# ####################################################################################################

# predict_inputs = pd.DataFrame({ 'mrna_ic':[1]*len(response_names),
#                                 'prot_ic':[1]*len(response_names),
#                                 'cntrl_1':[0.1]*len(response_names),
#                                 'cntrl_2':[1.0]*len(response_names),
#                                 'cntrl_3':[0.1]*len(response_names),
#                                 'Variable':response_names})

# true_pars = [np.log(0.5),np.log(1.1),np.log(2.1),np.log(0.3)]
# predictions = ode_model.predict(predict_inputs,true_pars)

# digit_re=re.compile('[a-z]+_t(\d+)')
# type_re=re.compile('([a-z]+)_t\d+')

# predictions['Prediction','Time'] = predictions['Inputs','Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
# predictions['Prediction','Type'] = predictions['Inputs','Variable'].apply(lambda x: type_re.search(x).group(1))

# mrna_pred = predictions.loc[predictions['Prediction','Type'] == 'mrna']
# prot_pred = predictions.loc[predictions['Prediction','Type'] == 'prot']

# ax1 = mrna_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red')
# ax2 = prot_pred.plot.scatter(x=('Prediction','Time'),y=('Prediction','Mean'),c='Red')

# ode_data = ode_model.sample(design,true_pars,design_replicats=3)

# data=cp.deepcopy(ode_data[0])

# data['Time'] = data['Variable'].apply(lambda x: int(digit_re.search(x).group(1)))
# data['Type'] = data['Variable'].apply(lambda x: type_re.search(x).group(1))

# mrna_obs = data.loc[data['Type'] == 'mrna']
# prot_obs = data.loc[data['Type'] == 'prot']

# mrna_obs.plot.scatter(x=('Time'),y=('Observation'),ax=ax1)
# prot_obs.plot.scatter(x=('Time'),y=('Observation'),ax=ax2)
# plt.show()


# fit_options={'Confidence':'Intervals',
#             'InitParamBounds':[(-3,2)]*4,
#             'InitSearchNumber':3,
#             'MaxSteps':100000}
# ode_fit = ode_model.fit(ode_data, options=fit_options)


# opts={'Covariance':True,'Bias':True,'MSE':True}
# eval_dat = ode_model.evaluate(design,true_pars,opts)

# t=0

# # fit_pars = mixed_fit['Estimate'].to_numpy().flatten()
# # print(fit_pars)

# # #evaluate goes here

# # predict_inputs = pd.DataFrame({ 'x1':[-1,-1,-1,0,0,0,1,1,1]*3,
# #                                 'x2':[-1,0,1,-1,0,1,-1,0,1]*3,
# #                                 'Variable':['y_norm']*9+['y_bern']*9+['y_pois']*9})

# # cov_mat = np.diag([0.1,0.1,0.1,0.1])
# # pred_options = {'Method':'MonteCarlo',
# #                 'PredictionInterval':True,
# #                 'ObservationInterval':True,
# #                 'Sensitivity':True}
# # predictions_dlta = mixed_model.predict(predict_inputs,fit_pars,covariance_matrix = cov_mat,options=pred_options)





""" 
Add a docstring
"""
import numpy as np
import pandas as pd
import casadi as cs
import matplotlib.pyplot as plt
from nloed import Model
from nloed import Design
####################################################################################################
# Block 1: Casadi RHS setup
####################################################################################################
# state variales
states = cs.SX.sym('states',2)
# control inputs
inducer = cs.SX.sym('inducer')
# parameters
parameters = cs.SX.sym('parameters',6)
#log-transform the parameters
alpha = cs.exp(parameters[0]) 
K = cs.exp(parameters[1]) 
delta = cs.exp(parameters[2])
beta = cs.exp(parameters[3]) 
L = cs.exp(parameters[4]) 
gamma = cs.exp(parameters[5])
#symbolic RHS
rhs = cs.vertcat(alpha*inducer/(K + inducer) - delta*states[0],
                 beta*states[0]/(L + states[0]) - gamma*states[1])
#casadi RHS function
rhs_func = cs.Function('rhs_func',[states,inducer,parameters],[rhs])

####################################################################################################
# Block 2: Integrator setup
####################################################################################################
#time step size
dt = 1
# Create symbolics for RK4 integration, as shown in Casadi examples
k1 = rhs_func(states, inducer, parameters)
k2 = rhs_func(states + dt/2.0*k1, inducer, parameters)
k3 = rhs_func(states + dt/2.0*k2, inducer, parameters)
k4 = rhs_func(states + dt*k3, inducer, parameters)
state_step = states + dt/6.0 * (k1 + 2*k2 + 2*k3 + k4)
# Create a function to perform one step of the RK integration
step_func = cs.Function('step_func',[states, inducer, parameters],[state_step])

####################################################################################################
# Block 3: define sampling and control intervals
####################################################################################################
#  cntrl_int1    cntrl_int2    cntrl_int3
#|------------||------------||------------|
#|-1---2-----3||-1---2-----3||-1---2-----3|
#repeated sampling pattern (1+2+3=6 steps long)
#total of 3*6
#set number of control intervals
num_cntrl_intervals = 3
#define a sample pattern to apply in each control interval
sample_pattern= [1,2,3] 
# define initial steady state input for ICs
initial_inducer = cs.SX.sym('init_inducer')
#define the steady-state initial states in terms of the initial inducer
init_mrna = (alpha/delta)*initial_inducer/(K+initial_inducer)
ini_prot = (beta/gamma)*init_mrna/(L+init_mrna)
# zip the initial states into a vector
initial_states = cs.vertcat(init_mrna,ini_prot)
# control values (one per control interval)
inducer_vector = cs.SX.sym('inducer_vec',3)
# experimental inputs
inputs = cs.vertcat(initial_inducer,inducer_vector)
# create list to store symbols for each sample point
sample_list=[]
# create list to store times of each sample (for naming observ. vars)
times=[]
#Loop over the ste function to create symbolics for integrating across a sample interval
state_sample = initial_states
step_counter=0
#loop over control invervals
for interval in range(num_cntrl_intervals):
  # loop over sample pattern
  for num_stps in sample_pattern:
    #iterate steps indicated by pattern
    for k in range(num_stps):
      #propogate the state variables
      state_sample = step_func(state_sample, inducer_vector[interval], parameters)
      step_counter+=1
    #save the state symbols and times of each sample
    sample_list.append(state_sample)
    times.append(step_counter*dt)

####################################################################################################
# Block 2: Model creation
####################################################################################################
# create list for observation structure 
observation_structure = []
#create list to store response names
observation_names, observation_type, observation_times = [], [], []
# loop over samples (time points)
for i in range(len(sample_list)):
  #create a unique name for mrna and prot samples
  mrna_name = 'mrna_'+'t'+"{0:0=2d}".format(times[i])
  prot_name = 'prot_'+'t'+"{0:0=2d}".format(times[i])
  #create mean and var tuple for mrna and prot observ.
  mrna_stats = cs.vertcat(sample_list[i][0], (0.1*sample_list[i][0])**2)
  prot_stats = cs.vertcat(sample_list[i][1], (0.1*sample_list[i][1])**2)
  #create casadi function for mrna and prot stats
  mrna_func = cs.Function(mrna_name,[inputs,parameters],[mrna_stats])
  prot_func = cs.Function(prot_name,[inputs,parameters],[prot_stats])
  #append the casadi function and distribution type to obs struct
  observation_structure.extend([(mrna_func,'Normal'),(prot_func,'Normal')])
  #store observation names, useful for plotting
  observation_names.extend([mrna_name,prot_name])
  #store observation type
  observation_type.extend(['RNA','Prot'])
  #store observation time
  observation_times.extend([times[i]]*2)

input_names = ['Init_Inducer','Inducer_1','Inducer_2','Inducer_3']
parameter_names = ['log_Alpha','log_K','log_Delta','log_Beta','log_L','log_Gamma']

model_object = Model(observation_structure, input_names, parameter_names)

#***hidden_start****
#generate initial dataset 
# create data frame of inputs that need predictions
init_design = pd.DataFrame({ 'Init_Inducer':[0.]*len(observation_names),
                                'Inducer_1':[1.]*len(observation_names),
                                'Inducer_2':[0.]*len(observation_names),
                                'Inducer_3':[3.]*len(observation_names),
                                'Variable':observation_names,
                                'Replicats':[1]*len(observation_names)})
true_param = np.log([2,1.5,1,3,0.75,0.5])
init_data = model_object.sample(init_design,true_param)
print('')
print(init_data)
print('')
#***hidden_end****

####################################################################################################
# Block 3: Inital data and fit
####################################################################################################
#request contours and use a simple initial search
fit_options={'Confidence':'None',
             'InitParamBounds':[(-1,2),(-1,2),(-1,2),(-1,2),(-1,2),(-1,2)],
             'InitSearchNumber':7,
             'SearchBound':5.}
#fit the model to the initial data
fit_info = model_object.fit(init_data, options=fit_options)
#extract the parameter values
fit_params = fit_info['Estimate'].to_numpy().flatten()
#***hidden_start****
print('')
print(np.exp(fit_params))
print('')
#***hidden_end****

####################################################################################################
# Block 4: CI comparison for initial data
####################################################################################################
# #***hidden_start****
# lik_bnds_array = np.hstack((fit_info['Lower'].to_numpy().T,
#                         fit_info['Estimate'].to_numpy().T,
#                         fit_info['Upper'].to_numpy().T))
# lik_bnds = pd.DataFrame(np.exp(lik_bnds_array),
#                         index=['Alpha0','Alpha','n','K'],
#                         columns=['Lower','Estimate','Upper'])
# print('')
# print(lik_bnds)
# print('')
# #***hidden_end****
#request the asymptotic covariance matrix
eval_options={'Method':'Asymptotic',
              'Covariance':True}
# call evaluate to compute the asymptotic covariance
asymptotic_covariance = model_object.evaluate(init_design,fit_params,eval_options)
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds0 = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha','K','Delta','Beta','L','Gamma'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds0)
print('')
# #***hidden_end****
# #generate 500 simulated datasets
# mc_data = model_object.sample(init_design,fit_params,design_replicats=500)
# #use the same pre-search settings as for the initial fit
# # fit_options={'InitParamBounds':[(-1,2),(1,3),(-1,2),(-1,2)],
# #              'InitSearchNumber':7}
# #fit the 500 datasets
# mc_fits = model_object.fit(mc_data, start_param = fit_params)
# #compute the Monte Carlo covariance estimate
# mc_covariance = np.cov(np.exp(mc_fits['Estimate'].to_numpy().T))
# #compute the asymptotic upper and lower 95% bounds
# mc_lower_bound = fit_params - 2*np.sqrt(np.diag(mc_covariance))
# mc_upper_bound = fit_params + 2*np.sqrt(np.diag(mc_covariance))
# #***hidden_start****
# mc_bnds_array = np.vstack((mc_lower_bound,
#                            fit_params,
#                            mc_upper_bound)).T
# mc_bnds = pd.DataFrame(np.exp(mc_bnds_array),
#                         index=['Alpha0','Alpha','n','K'],
#                         columns=['Lower','Estimate','Upper'])
# print('')
# print(mc_bnds)
# print('')
# #***hidden_end****

####################################################################################################
# Block 5: plot model predictions and initial data
####################################################################################################
#convert the covariance matrix to numpy
covariance_matrix = asymptotic_covariance.to_numpy()
#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({'Init_Inducer':[0.]*len(observation_names),
                                  'Inducer_1':[1.]*len(observation_names),
                                  'Inducer_2':[0.]*len(observation_names),
                                  'Inducer_3':[3.]*len(observation_names),
                                  'Variable':observation_names})
#request prediction and observation intervals
prediction_options = {'PredictionInterval':True,
                      'ObservationInterval':True}
#generate predictions and intervals
predictions = model_object.predict(prediction_inputs,
                                   fit_params,
                                   covariance_matrix = covariance_matrix,
                                   options=prediction_options)
#***hidden_start****
#partition the predictions by state type
mrna_pred = predictions.loc[[obs == 'RNA' for obs in observation_type]]
prot_pred = predictions.loc[[obs == 'Prot' for obs in observation_type]]
#partition the initial data predictions by state type
mrna_dat = init_data.loc[[obs == 'RNA' for obs in observation_type]]
prot_dat = init_data.loc[[obs == 'Prot' for obs in observation_type]]
#plot the predictions, intervals and data, two subplots; mrna and protein
fig, ax = plt.subplots(2)
obs_bars = [(mrna_pred['Prediction','Mean']  - mrna_pred['Observation','Lower']).to_list(),
            (mrna_pred['Observation','Upper'] - mrna_pred['Prediction','Mean']).to_list()]
ax[0].errorbar(times, mrna_pred['Prediction','Mean'], yerr=obs_bars, fmt='none', color='C1')
pred_bars = [(mrna_pred['Prediction','Mean']  - mrna_pred['Prediction','Lower']).to_list(),
             (mrna_pred['Prediction','Upper'] - mrna_pred['Prediction','Mean']).to_list()]
ax[0].errorbar(times, mrna_pred['Prediction','Mean'], yerr=pred_bars, fmt='.', color='C0')
ax[0].plot(times, mrna_dat['Observation'], 'x', color='C1')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('mRNA')
obs_bars = [(prot_pred['Prediction','Mean']  - prot_pred['Observation','Lower']).to_list(),
            (prot_pred['Observation','Upper'] - prot_pred['Prediction','Mean']).to_list()]
ax[1].errorbar(times, prot_pred['Prediction','Mean'], yerr=obs_bars, fmt='none', color='C1')
pred_bars = [(prot_pred['Prediction','Mean']  - prot_pred['Prediction','Lower']).to_list(),
             (prot_pred['Prediction','Upper'] - prot_pred['Prediction','Mean']).to_list()]
ax[1].errorbar(times, prot_pred['Prediction','Mean'], yerr=pred_bars, fmt='.', color='C0')
ax[1].plot(times, prot_dat['Observation'], 'x', color='C1')
ax[1].set_xlabel('Time')
ax[1].set_ylabel('Prot')
plt.show()
# #***hidden_end****

####################################################################################################
# Block 6: optimal design
####################################################################################################
#set up the design algorithm to use continuous (continuous) optimization with two unique inputs points
continuous_inputs={'Inputs':['Init_Inducer','Inducer_1','Inducer_2','Inducer_3'],
                   'Bounds':[(.01,5),(.01,5),(.01,5),(.01,5)],
                   'Structure':[['I0','I1','I2','I3']]}
#set up fixed design dictionary
fixed_dict ={'Weight':0.5, 'Design':init_design}
# generate the optimal discreteimate (relaxed) design
design_object = Design(model_object,fit_params,'D',
                       fixed_design = fixed_dict,
                       continuous_inputs = continuous_inputs)
#extract the relaxed design structure
relaxed_design = design_object.relaxed()
#***hidden_start****
print('')
print(relaxed_design)
print('')
#***hidden_end****
#set the sample size to 30
sample_size = 18
#generate a rounded exact design 
exact_design = design_object.round(sample_size)
#***hidden_start****
print('')
print(exact_design)
print('')
#***hidden_end****

####################################################################################################
# Block 7: fit combined data
####################################################################################################
#***hidden_start****
#generate some data for optimal design and combine 
optimal_data = model_object.sample(exact_design,true_param)
combined_data = pd.concat([init_data, optimal_data], ignore_index=True)
#***hidden_end****
#combine the initial and optimal design
combined_design = pd.concat([init_design, exact_design], ignore_index=True)
#request contours and use a simple initial search
fit_options={'Confidence':'None'}
#fit the model to the initial data
fit_info = model_object.fit(combined_data,start_param = fit_params, options=fit_options)
#extract the parameter values
fit_params = fit_info['Estimate'].to_numpy().flatten()
#***hidden_start****
print('')
print(np.exp(fit_params))
print('')
#***hidden_end****

####################################################################################################
# Block 8: CI comparison for initial data
####################################################################################################
#***hidden_start****
# lik_bnds_array = np.hstack((fit_info['Lower'].to_numpy().T,
#                         fit_info['Estimate'].to_numpy().T,
#                         fit_info['Upper'].to_numpy().T))
# lik_bnds = pd.DataFrame(np.exp(lik_bnds_array),
#                         index=['Alpha0','Alpha','n','K'],
#                         columns=['Lower','Estimate','Upper'])
# print('')
# print(lik_bnds)
# print('')
#***hidden_end****
#request the asymptotic covariance matrix
eval_options={'Method':'Asymptotic',
              'Covariance':True}
# call evaluate to compute the asymptotic covariance
asymptotic_covariance = model_object.evaluate(combined_design,fit_params,eval_options)
#compute the asymptotic upper and lower 95% bounds
asymptotic_lower_bound = fit_params - 2*np.sqrt(np.diag(asymptotic_covariance))
asymptotic_upper_bound = fit_params + 2*np.sqrt(np.diag(asymptotic_covariance))
#***hidden_start****
fim_bnds_array = np.vstack((asymptotic_lower_bound,
                            fit_params,
                            asymptotic_upper_bound)).T
fim_bnds = pd.DataFrame(np.exp(fim_bnds_array),
                        index=['Alpha','K','Delta','Beta','L','Gamma'],
                        columns=['Lower','Estimate','Upper'])
print('')
print(fim_bnds)
print('')
# #***hidden_end****
# #generate 500 simulated datasets
# mc_data = model_object.sample(init_design,fit_params,design_replicats=500)
# #fit the 500 datasets
# mc_fits = model_object.fit(mc_data, start_param = fit_params)
# #compute the Monte Carlo covariance estimate
# mc_covariance = np.cov(np.exp(mc_fits['Estimate'].to_numpy().T))
# #compute the asymptotic upper and lower 95% bounds
# mc_lower_bound = fit_params - 2*np.sqrt(np.diag(mc_covariance))
# mc_upper_bound = fit_params + 2*np.sqrt(np.diag(mc_covariance))
# #***hidden_start****
# mc_bnds_array = np.vstack((mc_lower_bound,
#                            fit_params,
#                            mc_upper_bound)).T
# mc_bnds = pd.DataFrame(np.exp(mc_bnds_array),
#                         index=['Alpha0','Alpha','n','K'],
#                         columns=['Lower','Estimate','Upper'])
# print('')
# print(mc_bnds)
# print('')
# #***hidden_end****

####################################################################################################
# Block 5: plot model predictions and initial data
####################################################################################################
#convert the covariance matrix to numpy
covariance_matrix = asymptotic_covariance.to_numpy()
#generate predictions with error bars fdor a random selection of inputs)
prediction_inputs = pd.DataFrame({'Init_Inducer':[0.]*len(observation_names),
                                  'Inducer_1':[1.]*len(observation_names),
                                  'Inducer_2':[0.]*len(observation_names),
                                  'Inducer_3':[3.]*len(observation_names),
                                  'Variable':observation_names})
#request prediction and observation intervals
prediction_options = {'PredictionInterval':True,
                      'ObservationInterval':True}
#generate predictions and intervals
predictions = model_object.predict(prediction_inputs,
                                   fit_params,
                                   covariance_matrix = covariance_matrix,
                                   options=prediction_options)
#***hidden_start****
#partition the predictions by state type
mrna_pred = predictions.loc[[obs == 'RNA' for obs in observation_type]]
prot_pred = predictions.loc[[obs == 'Prot' for obs in observation_type]]
#partition the initial data predictions by state type
mrna_dat = init_data.loc[[obs == 'RNA' for obs in observation_type]]
prot_dat = init_data.loc[[obs == 'Prot' for obs in observation_type]]
#plot the predictions, intervals and data, two subplots; mrna and protein
fig, ax = plt.subplots(2)
obs_bars = [(mrna_pred['Prediction','Mean']  - mrna_pred['Observation','Lower']).to_list(),
            (mrna_pred['Observation','Upper'] - mrna_pred['Prediction','Mean']).to_list()]
ax[0].errorbar(times, mrna_pred['Prediction','Mean'], yerr=obs_bars, fmt='none', color='C1')
pred_bars = [(mrna_pred['Prediction','Mean']  - mrna_pred['Prediction','Lower']).to_list(),
             (mrna_pred['Prediction','Upper'] - mrna_pred['Prediction','Mean']).to_list()]
ax[0].errorbar(times, mrna_pred['Prediction','Mean'], yerr=pred_bars, fmt='.', color='C0')
ax[0].plot(times, mrna_dat['Observation'], 'x', color='C1')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('mRNA')
obs_bars = [(prot_pred['Prediction','Mean']  - prot_pred['Observation','Lower']).to_list(),
            (prot_pred['Observation','Upper'] - prot_pred['Prediction','Mean']).to_list()]
ax[1].errorbar(times, prot_pred['Prediction','Mean'], yerr=obs_bars, fmt='none', color='C1')
pred_bars = [(prot_pred['Prediction','Mean']  - prot_pred['Prediction','Lower']).to_list(),
             (prot_pred['Prediction','Upper'] - prot_pred['Prediction','Mean']).to_list()]
ax[1].errorbar(times, prot_pred['Prediction','Mean'], yerr=pred_bars, fmt='.', color='C0')
ax[1].plot(times, prot_dat['Observation'], 'x', color='C1')
ax[1].set_xlabel('Time')
ax[1].set_ylabel('Prot')
plt.show()
# #***hidden_end****


#!/usr/bin/env python
# coding: utf-8

# # A Simple 2-Input Linear Regression Model
# 
# This example explains how to create a simple 2-input linear regression model using NLoed's Model class.

# ### Create Casadi Symbols for the Model
# 
# In this example, we assume there is a linear relationship between two input variables the experimenter controls: $x_1$ and $x_2$, and the observation variable $y$. This relationship can be described by the equation linear equation:
# 
# $\hat{y} = \beta_{0} + \beta_{1} x_{1} + \beta_{2} x_{2}$
# 
# Here $\hat{y}$ represents the mean of the observation response. In this example we assum that $y$ follows a normal distribution with constant known variance of Var($y$)$=1$.
# 
# In order to implement this model in NLoed's Model class we start by importing the NLoed and Casadi packages. We then create creat symbols for the inputs, $x_i$, and the parameters $\beta_j$. This can be done as follows:

# In[1]:


import nloed as nl
import casadi as cs
x = cs.SX.sym('X',2)
beta = cs.SX.sym('beta',3)


# Now, we use the Casadi symbols to implement an expression for the deterministic model equation. This expression links the mean of the obeservation variable $y$ to the inputs and the parameters. This is done as follows:

# In[2]:


y_mean = beta[0] + beta[1]*x[0] + beta[2]*x[1] 


# The variance of the obervation variable, $y$ is a constant, with no relation to the model inputs or the parameters. We can therefore set it as constant in our model preparation.

# In[3]:


y_var = 0.1


# ### Instantiating an NLoed Model Object
# 
# Model class's constructor requires the model to be encoded as a Casadi function, mapping the model inputs and parameters to the sampling statistics of the observation variable. In this case, as we assume $y$ has a normal distribution, it's sampling statistics consist of the mean and variance of $y$. To create a Casadi function for this relationship, we zip the the mean and variance into a single vector and call Casadi's function constructor as follows:

# In[4]:


y_stats = cs.vertcat(y_mean, y_var)
y_func = cs.Function('Observable_y',[x,beta],[y_stats])


# Above, we passed a string 'Observable_y' to the function constructor to give the Casadi function a name attribute. This name attribute will become the observation's name in the NLoed Model instance. Casadi's constructor also requires us to provide names for the inputs and parameters. This is done usign two lists of strings as follows:

# In[5]:


input_names = ['Input_x1','Input_x2']
param_names = ['Intercept','Slope1','Slope2']


# Finally, the NLoed model constructor needs to know that the observation variable $y$ is normally distributed. To label it as such we tuple the Casadi function, *y_func*, encoding the mdoel with the string, "Normal". This tuple is then places in a list, *observ_list*. If the model had other observation variables that also depended on the same inputs and parameters, they could be added to this list as additional tuples each with their own distribution. However, this model only contains one output observable so the *observ_list* looks like this:

# In[6]:


observ_list = [(y_func,'Normal')]


# And at last, the NLoed model is declared using the constructor as follows:

# In[7]:


linear_model = nl.Model(observ_list,input_names,param_names)


# The *linear_model* object can now be used for many things. Check out some of the other notebooks for examples.
